/// <reference path="./modules.d.ts" />
import './redirect';
import type { Rule } from 'eslint';
import type { ESLintUtils } from '@typescript-eslint/utils';
import type { ESLintConfig, Extends, Plugin, Rules } from '@aet/eslint-define-config';
// import findCacheDirectory from 'find-cache-dir';
import { typescriptRules } from './presets/typescript';
import { unicornRules } from './presets/unicorn';
import { eslintRules } from './presets/eslint';
import { reactRules } from './presets/react';
import { importRules } from './presets/import-x';
import { jsDocRules } from './presets/jsdoc';
import { graphqlRules } from './presets/graphql';
import { localRules } from './presets/local';
import { error, warn, off } from './constants';
import { tailwindRules } from './presets/tailwind';
import { checkEnv } from './env';

export { error, warn, off };

declare global {
  interface Array<T> {
    filter(
      predicate: BooleanConstructor,
    ): Exclude<T, null | undefined | false | '' | 0>[];
  }
}

const unique = (...arr: (false | undefined | string | string[])[]): string[] => [
  ...new Set(arr.flat(1).filter(Boolean)),
];

const ensureArray = <T>(value?: T | T[]): T[] =>
  value == null ? [] : Array.isArray(value) ? value : [value];

type RuleLevel = 'error' | 'warn' | 'off' | 0 | 1 | 2;
type RuleEntry<Options> = RuleLevel | [RuleLevel, Partial<Options>];

export interface LocalRuleOptions {
  /** Bans import from the specifier '.' and '..' and replaces it with '.+/index' */
  'rules/no-import-dot': RuleEntry<unknown>;
  /**
   * Enforce template literal expressions to be of `string` type
   * @see [restrict-template-expressions](https://typescript-eslint.io/rules/restrict-template-expressions)
   */
  'rules/restrict-template-expressions': RuleEntry<{ allow: string[] }>;
  /** Ban assignment of empty object literals `{}` and replace them with `Object.create(null)` */
  'rules/no-empty-object-literal': RuleEntry<unknown>;
}

export type RuleOptions = Rules & Partial<LocalRuleOptions>;

export interface CustomRule {
  rule: () => Promise<{
    default: Rule.RuleModule | ESLintUtils.RuleModule<string, unknown[]>;
  }>;
  options?: RuleLevel;
}

/**
 * ESLint Configuration.
 * @see [ESLint Configuration](https://eslint.org/docs/latest/user-guide/configuring/)
 */
export type InputConfig = Omit<ESLintConfig, 'rules'> & {
  /**
   * Rules.
   * @see [Rules](https://eslint.org/docs/latest/user-guide/configuring/rules)
   */
  rules?: Partial<RuleOptions>;

  /**
   * Glob pattern to find paths to custom rule files in JavaScript or TypeScript.
   * Note this must be a string literal or an array of string literals since
   * this is statically analyzed.
   */
  customRuleFiles?: string | string[];

  /**
   * Automatically detect project types, dependencies and deduct the plugins.
   */
  auto?: boolean;
};

/**
 * Returns a ESLint config object.
 *
 * By default, it includes `["@typescript-eslint", "import-x", "prettier", "unicorn"]` configs.
 * Additional bundled plugins include:
 *
 * 1. [`react`](https://github.com/jsx-eslint/eslint-plugin-react#list-of-supported-rules)
 *    (automatically enables
 *    [`react-hooks`](https://github.com/facebook/react/tree/main/packages/eslint-plugin-react-hooks))
 * 2. [`react-refresh`](https://github.com/ArnaudBarre/eslint-plugin-react-refresh)
 * 3. [`jsx-a11y`](https://github.com/jsx-eslint/eslint-plugin-jsx-a11y#supported-rules)
 * 4. [`unicorn`](https://github.com/sindresorhus/eslint-plugin-unicorn#rules)
 * 5. [`n`](https://github.com/eslint-community/eslint-plugin-n#-rules) (Node.js specific,
 *    requires `minimatch`)
 * 6. [`jsdoc`](https://github.com/gajus/eslint-plugin-jsdoc#rules)
 *
 * Non bundled:
 * 1. [`graphql`](https://the-guild.dev/graphql/eslint/rules)
 */
export function extendConfig(of: InputConfig = {}): ESLintConfig {
  const {
    auto,
    plugins: _plugins = [],
    settings,
    rules,
    extends: _extends,
    overrides,
    customRuleFiles,
    parserOptions,
    // @ts-expect-error
    localRules: _,
    ...rest
  } = of;

  let hasReact = false;
  let plugins: Plugin[] = _plugins;
  let extend: Extends[] = ensureArray(_extends);

  if (auto) {
    const env = checkEnv();
    hasReact = env.react;
    plugins = [..._plugins, ...env.plugins];
    extend = [...extend, ...env.extends];
  }

  const hasJsDoc = plugins.includes('jsdoc');
  const hasGraphQL = plugins.includes('@graphql-eslint');
  const hasTailwind = extend.some(name => name.includes('plugin:tailwindcss/'));

  // const ruleDir = false; // ?? findCacheDirectory({ name: '_eslint-rules' });
  // if (ruleDir) {
  //   fs.rmSync(ruleDir, { recursive: true, force: true });
  //   fs.mkdirSync(ruleDir, { recursive: true });
  // }

  const result: InputConfig = {
    root: true,
    parser: '@typescript-eslint/parser',
    plugins: unique('@typescript-eslint', 'import-x', 'rules', 'unicorn', plugins),
    env: { node: true, browser: true, es2023: true },
    reportUnusedDisableDirectives: true,
    parserOptions: {
      project: true,
      ...parserOptions,
    },
    extends: unique(
      'eslint:recommended',
      'prettier',
      'plugin:@typescript-eslint/recommended-type-checked',
      'plugin:import-x/errors',
      'plugin:import-x/typescript',
      hasReact && [
        'plugin:@eslint-react/recommended-legacy',
        'plugin:@eslint-react/dom-legacy',
        'plugin:react-hooks/recommended',
        'plugin:jsx-a11y/recommended',
      ],
      hasJsDoc && 'plugin:jsdoc/recommended-typescript',
      hasGraphQL && 'plugin:@graphql-eslint/recommended',
      extend as string[],
    ),
    settings: {
      'import-x/parsers': {
        '@typescript-eslint/parser': ['.ts', '.tsx', '.mts', '.cts'],
      },
      'import-x/resolver': {
        typescript: {
          alwaysTryTypes: true,
        },
      },
      ...settings,
    },
    overrides: [
      {
        files: ['.eslintrc.js', '.eslintrc.cjs', '*.config.js', 'index.js'],
        extends: ['plugin:@typescript-eslint/disable-type-checked'],
        rules: {
          'rules/restrict-template-expressions': off,
        },
      },
      {
        files: ['*.d.ts'],
        rules: {
          '@typescript-eslint/consistent-type-imports': off,
        },
      },
      {
        files: ['repl.ts', 'scripts/**/*.ts'],
        rules: {
          'no-console': off,
        },
      },
      {
        files: ['*.tsx'],
        rules: {
          '@eslint-react/no-leaked-conditional-rendering': error,
        },
      },
      ...(overrides ?? []),
    ],
    rules: {
      ...eslintRules,
      ...typescriptRules,
      ...importRules,
      ...localRules,
      ...(hasReact && reactRules),
      ...(plugins.includes('react-refresh') && {
        'react-refresh/only-export-components': [warn, { allowConstantExport: true }],
      }),
      ...unicornRules,
      ...(hasJsDoc && jsDocRules),
      ...(hasGraphQL && graphqlRules),
      ...(hasTailwind && tailwindRules),
      ...rules,
    },
    ...rest,
  };

  return result;
}
